@@ -1,6 +1,6 @@
 /*
 integrit - file integrity verification system
-Copyright (C) 2001 Ed Cashin
+Copyright (C) 2000, 2001 Ed L. Cashin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -22,7 +22,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 typedef struct cdb_seq {
   int		fd;
-  size_t	eod;		/* end of data */
+  uint32_t	eod;		/* end of data */
 } cdb_seq;
 
 int cdb_seq_start(cdb_seq *c);
