@@ -1,6 +1,6 @@
 /*
 integrit - file integrity verification system
-Copyright (C) 2001 Ed Cashin
+Copyright (C) 2000, 2001 Ed L. Cashin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -24,6 +24,14 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include	<unistd.h>
 #include	<ctype.h>
 #include	<errno.h>
+/* support platforms that don't yet conform to C99 */
+#if	HAVE_STDINT_H
+#include	<stdint.h>
+#elif	HAVE_INTTYPES_H
+#include	<inttypes.h>
+#else
+#error No stdint.h or inttypes.h found.
+#endif
 #include	"cdb.h"
 #include	"cdb_make.h"
 #include	"hashtbl.h"
@@ -41,10 +49,13 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include	"leakfind.h"
 #endif
 
+/* There is a thread on comp.lang.c.moderated where the gurus explain
+ * how the cast on the parameters to ctype routines should be to
+ * unsigned char. */
 #define	EATSPACE(buf)	do {				\
     while (*(buf)					\
 	   && (*(buf) != '\n')				\
-	   && (isspace((int) *(buf))))	/* skip whitespace */	\
+	   && (isspace((unsigned char) *(buf))))	\
       ++(buf);						\
 } while (0)    
 
@@ -105,7 +116,8 @@ void options_announce(FILE *out, options *o)
 {
     switch (o->output) {
       case OUTPUT_LINES:
-	options_announce_lines(o, out);
+	if (o->verbose > 0)
+	  options_announce_lines(o, out);
 	break;
       case OUTPUT_XML:
 	options_announce_xml(o, out);
@@ -135,12 +147,18 @@ void options_init(options *o)
     o->output		 = OUTPUT_LINES; /* human-readable output */
 }
 
+void *call_checkset_free(void *cset, void *data)
+{
+    checkset_free((checkset *) cset);
+    return NULL;
+}
+
 void options_destroy(options *o)
 {
     free(o->knowndbname);
     free(o->currdbname);
     free(o->root);
-    hashtbl_free(o->ruleset);
+    hashtbl_free(o->ruleset, call_checkset_free, NULL);
     hashtbl_destroy(o->ruleset);
     free(o->ruleset);
 }
@@ -178,12 +196,22 @@ inline static void do_assignment(options *o, char *buf, char *eq)
 
     *eq	 = '\0';
     if ( (property = strstr(buf, "known")) ) {
+      /* the commandline overrides the config file
+       * (really the first one to set it wins and the rest are ignored)
+       */
+      if (o->knowndbname)
+	return;
       EATSPACE(val);
       if (! *val)
 	die_noprop(property, __FUNCTION__);
       chomp(val);
       o->knowndbname	 = xstrdup(val);
     } else if ( (property = strstr(buf, "current")) ) {
+      /* the commandline overrides the config file
+       * (really the first one to set it wins and the rest are ignored)
+       */
+      if (o->currdbname)
+	return;
       EATSPACE(val);
       if (! *val)
 	die_noprop(property, __FUNCTION__);
@@ -201,28 +229,31 @@ inline static void do_assignment(options *o, char *buf, char *eq)
 }
 
 static void options_add_checkset(options *o, char *namebuf,
-				 size_t namebuf_chars, checkset cset)
+				 size_t namebuf_chars, checkset *cset)
 {
     hashtbl_t	*h	 = o->ruleset;
-    checkset	old;
+    checkset	*old;
+
+    if ( (old = hashtbl_lookup(h, namebuf, namebuf_chars)) )
+      cset	 = checkset_merge(cset, old, namebuf);
 
-    if ( (old = hashtbl_lookup(h, namebuf, namebuf_chars)) ) {
-      fprintf(stderr, "Warning: overwriting old checkset (");
-      checkset_show(stderr, old);
-      fprintf(stderr, ") for file (%s)\n", namebuf);
-      free(old);
-    }
     hashtbl_store(h, namebuf, namebuf_chars, cset);
 }
 
+/* this function is made more complex by the fact that is supports both
+ * non-inheriting and (regular) cascading checksets.  The cset pointer
+ * is overloaded to point to a non-inheriting checkset (ni_checkset)
+ * when necessary.
+ */
 inline static void do_rule(options *o, char *buf)
 {
-    checkset	cset;
+    checkset	*cset;
     char	namebuf[BUFSIZ];
     int		namebuf_chars	 = 0;
     int		n_switches;
     char	ignore		 = 0;
     char	nochildren	 = 0;
+    char	noinherit	 = 0;
 
     EATSPACE(buf);
     switch (*buf) {
@@ -234,6 +265,10 @@ inline static void do_rule(options *o, char *buf)
 	nochildren	 = 1;
 	++buf;
 	break;
+      case '$':
+	noinherit	 = 1;
+	++buf;
+	break;
       default:
 	break;	
     }
@@ -245,11 +280,11 @@ inline static void do_rule(options *o, char *buf)
 	die(__FUNCTION__, "Error: filename too long in config");
 
       if (*buf == '\\') {
-	if (isspace((int) *(buf + 1)))
+	if (isspace((unsigned char) *(buf + 1)))
 	  namebuf[namebuf_chars++]	 = *(++buf);
 	else
 	  namebuf[namebuf_chars++]	 = '\\';
-      } else if (isspace((int) *buf)) {
+      } else if (isspace((unsigned char) *buf)) {
 	break;			/* end of the name */
       } else {
 	namebuf[namebuf_chars++]	 = *buf;
@@ -260,7 +295,7 @@ inline static void do_rule(options *o, char *buf)
     EATSPACE(buf);
     n_switches	 = strlen(buf);
     /* trim trailing whitespace */
-    while (n_switches && isspace((int) buf[n_switches - 1])) {
+    while (n_switches && isspace((unsigned char) buf[n_switches - 1])) {
       buf[n_switches - 1]	 = '\0';
       --n_switches;
     }
@@ -269,18 +304,24 @@ inline static void do_rule(options *o, char *buf)
 	  "Error: unrecognized check switch in conf file rule for %s",
 	  namebuf);
 
-    if (! (cset = malloc(CHECKSET_SIZEOF(n_switches))) ) 
-      DIE("malloc checkset");
-    checkset_init(cset, n_switches);
-    CHECKSET_SETIGNORE(cset, ignore);
-    CHECKSET_SETNOCHILD(cset, nochildren);
+    if (noinherit)
+      cset	 = (checkset *) ni_checkset_new();
+    else
+      cset	 = checkset_new();
 
-    memcpy(CHECKSET_SWITCHSTART(cset), buf, n_switches);
+    CHECKSET_SETBOOLEAN(cset, RULE_IGNORE, ignore);
+    CHECKSET_SETBOOLEAN(cset, RULE_NOCHILDREN, nochildren);
+    CHECKSET_SETBOOLEAN(cset, RULE_NOINHERIT, noinherit);
+
+    if (noinherit)
+      ((ni_checkset *) cset)->ni_switches	 = xstrdup(buf);
+    else
+      cset->switches	 = xstrdup(buf);
     
     chomp(namebuf);
 #ifdef	DEBUG
     fprintf(stderr, "debug: path (%s) checkset (", namebuf);
-    checkset_show(stderr, cset);	/* debug */
+    checkset_show(stderr, cset);
     fputs(")\n", stderr);
 #endif
     
@@ -289,18 +330,22 @@ inline static void do_rule(options *o, char *buf)
 
 static void usage(void)
 {
-    fputs("usage:\n"
+    fputs("Usage:\n"
 	  "\n"
-	  "      integrit -C conffile [-x] [-u] [-c]\n"
+	  "      integrit -C conffile [options]\n"
 	  "      integrit -V\n"
 	  "      integrit -h\n"
 	  "\n"
-	  "options:\n"
+	  "Options:\n"
 	  "\n"
 	  "      -C	specify configuration file\n"
 	  "      -x	use XML output instead of abbreviated output\n"
 	  "      -u	do update: create current state database\n"
 	  "      -c	do check: verify current state against known db\n"
+	  "      -q	lower verbosity\n"
+	  "      -v	raise verbosity\n"
+	  "      -N	specify the current (New) database, overriding conf file\n"
+	  "      -O	specify the known (Old) database, overriding conf file\n"
 	  "      -V	show integrit version info and exit\n"
 	  "      -h	show this help      \n\n", stderr);
 }
@@ -310,7 +355,7 @@ void parse_args(options *o, int argc, char *argv[])
     int		c;
     opterr	 = 0;
 
-    while ( (c = getopt(argc, argv, "hVvqC:cux")) != -1) {
+    while ( (c = getopt(argc, argv, "hVvqC:cuxN:O:")) != -1) {
       switch (c) {
 	case 'h':
 	  usage();
@@ -332,6 +377,12 @@ void parse_args(options *o, int argc, char *argv[])
 	case 'c':
 	  o->do_check	 = 1;
 	  break;
+	case 'N':
+	  o->currdbname	 = xstrdup(optarg);
+	  break;
+	case 'O':
+	  o->knowndbname = xstrdup(optarg);
+	  break;
 	case 'u':
 	  o->do_update	 = 1;
 	  break;
