@@ -1,6 +1,6 @@
 /*
 integrit - file integrity verification system
-Copyright (C) 2000 Ed Cashin
+Copyright (C) 2000, 2001 Ed L. Cashin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -20,6 +20,14 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include	<config.h>
 #include	<stdio.h>
 #include	<time.h>
+/* support platforms that don't yet conform to C99 */
+#if	HAVE_STDINT_H
+#include	<stdint.h>
+#elif	HAVE_INTTYPES_H
+#include	<inttypes.h>
+#else
+#error No stdint.h or inttypes.h found.
+#endif
 #include	"cdb.h"
 #include	"cdb_make.h"
 #include	"hashtbl/hashtbl.h"
